#include "script/common.lua"

function VecDist(a, b)
	return VecLength(VecSub(a, b))
end


function getTagParameter(entity, name, default)
	--+print("getTagParameter")
	local v = tonumber(GetTagValue(entity, name))
	if v then
		return v
	else
		return default
	end
end

function getTagParameter2(entity, name, default)
	--+--+print("getTagParameter2")
	local s = splitString(GetTagValue(entity, name), ",")
	if #s == 1 then
		local v = tonumber(s[1])
		if v then
			return v, v
		else
			return default, default
		end
	elseif #s == 2 then
		local v1 = tonumber(s[1])
		local v2 = tonumber(s[2])
		if v1 and v2 then
			return v1, v2
		else
			return default, default
		end
	else
		return default, default
	end
end

oneHitSniper = GetBool("savegame.mod.destructible-bot.oneHitSniper")
robotAttackOneHit = GetBool("savegame.mod.destructible-bot.robotAttackOneHit")
robotOutline = GetBool("savegame.mod.destructible-bot.robotOutline")
robotBetterSensors = GetBool("savegame.mod.destructible-bot.robotBetterSensors")
robotHealth = GetFloat("savegame.mod.destructible-bot.robotHealth")
oneHitEverything = GetBool("savegame.mod.destructible-bot.oneHitEverything")



pType = GetStringParam("type", "")
pSpeed = GetFloatParam("speed", 2)
pTurnSpeed = GetFloatParam("turnspeed", pSpeed)

config = {}
config.hasVision = false


	config.viewFov = 150
	config.maxSoundDist = 100.0
	config.lostVisibilityTimer = 5.0 --Time player is seen after losing visibility
	config.huntSpeedScale = 1.6
	config.sensorDist = 5.0
	config.aimTime = 5.0
	config.viewDistance = 20
	config.visibilityTimer = 0.3 --Time player must be seen to be identified as enemy (ideal condition)



config.canHearPlayer = false
config.canSeePlayer = false
config.patrol = false
config.sensorDist = 5.0
config.speed = pSpeed
config.turnSpeed = pTurnSpeed
config.huntPlayer = false
config.avoidPlayer = false
config.triggerAlarmWhenSeen = false
config.noweapon = false

config.outline = 20


config.aggressive = false
config.stepSound = "m"
config.practice = GetBool("savegame.mod.destructible-bot.peacefulmode")

PATH_NODE_TOLERANCE = 0.8

function configInit()
	--+--+print("configInit")
	local eye = FindLight("eye")
	local head = FindBody("head")
	config.patrol = FindLocation("patrol") ~= 0
	config.hasVision = eye ~= 0
	config.viewDistance = getTagParameter(eye, "viewdist", config.viewDistance)
	if robotBetterSensors then
		config.viewDistance = 100
	end
	config.viewFov = getTagParameter(eye, "viewfov", config.viewFov)
	config.maxSoundDist = getTagParameter(head, "heardist", config.maxSoundDist)
	if hasWord(pType, "investigate") then
		config.canHearPlayer = true
		config.canSeePlayer = true
	end
	if hasWord(pType, "chase") then
		config.canHearPlayer = true
		config.canSeePlayer = true
		config.huntPlayer = true
	end


	if hasWord(pType, "avoid") and config.patrol then
		config.avoidPlayer = true
		config.canSeePlayer = true
	end
	if hasWord(pType, "alarm") then
		config.triggerAlarmWhenSeen = true
	end
	if hasWord(pType, "nooutline") or false == GetBool("savegame.mod.destructible-bot.robotOutline") then
		config.outline = 0
	end
	if hasWord(pType, "aggressive") then
		config.aggressive = true
	end
	if hasWord(pType, "practice") then
		config.canSeePlayer = true
		config.practice = true
	end
	local body = FindBody("body")
	if HasTag(body, "stepsound") then
		config.stepSound = GetTagValue(body, "stepsound")
	end
end

------------------------------------------------------------------------


function rnd(mi, ma)
	--+--+print("rnd")
	local v = math.random(0,1000) / 1000
	return mi + (ma-mi)*v
end

function rejectAllBodies(bodies)
	--+--+print("rnd")
	for i=1, #bodies do
		QueryRejectBody(bodies[i])
	end
end

------------------------------------------------------------------------


robot = {}
robot.body = 0
robot.transform = Transform()
robot.axes = {}
robot.bodyCenter = Vec()
robot.navigationCenter = Vec()
robot.dir = Vec(0, 0, -1)
robot.speed = 0
robot.blocked = 0
robot.mass = 0
robot.allBodies = {}
robot.allShapes = {}
robot.allJoints = {}
robot.initialBodyTransforms = {}
robot.enabled = true
robot.deleted = false
robot.speedScale = 1
robot.breakAll = false
robot.breakAllTimer = 0
robot.dirToPlayer = 0
robot.distToPlayer = 100
robot.roamTrigger = 0
robot.limitTrigger = 0
robot.investigateTrigger = 0
robot.activateTrigger = 0
robot.stunned = 0
robot.outlineAlpha = 0
robot.canSensePlayer = false
robot.playerPos = Vec()



function robotSetAxes()
	--+--+print("robotSetAxes")
	robot.transform = GetBodyTransform(robot.body)
	robot.axes[1] = TransformToParentVec(robot.transform, Vec(1, 0, 0))
	robot.axes[2] = TransformToParentVec(robot.transform, Vec(0, 1, 0))
	robot.axes[3] = TransformToParentVec(robot.transform, Vec(0, 0, 1))
end


function robotInit()
	--+--+print("robotInit")
	robot.body = FindBody("body")
	robot.allBodies = FindBodies()
	robot.allShapes = FindShapes()
	robot.allJoints = FindJoints()
	robot.roamTrigger = FindTrigger("roam",true)
	robot.limitTrigger = FindTrigger("limit")
	robot.investigateTrigger = FindTrigger("investigate",true)
	robot.activateTrigger = FindTrigger("activate")
	if robot.activateTrigger ~= 0 then
		SetTag(robot.body, "inactive")
	end
	for i=1, #robot.allBodies do
		robot.initialBodyTransforms[i] = GetBodyTransform(robot.allBodies[i])
	end
	robotSetAxes()
end


function robotUpdate(dt)
	--+--+print("robotUpdate")
	robotSetAxes()

	if config.practice then
		local pp = GetPlayerCameraTransform().pos
		local pt = FindTrigger("practicearea")
		if pt ~= 0 and IsPointInTrigger(pt, pp) then
			robot.playerPos = VecCopy(pp)
			if not stackHas("navigate") then
				robotTurnTowards(robot.playerPos)
			end
		else
		local overrideTarget = FindBody("practicetarget", true)
		if overrideTarget ~= 0 then
			robot.playerPos = GetBodyTransform(overrideTarget).pos
			if not stackHas("navigate") then
				robotTurnTowards(robot.playerPos)
			end
		else
			robot.playerPos = Vec(0, -100, 0)
		end
	end
	else
		robot.playerPos = GetPlayerCameraTransform().pos
	end
	
	local vel = GetBodyVelocity(robot.body)
	local fwdSpeed = VecDot(vel, robot.dir)
	local blocked = 0
	if robot.speed > 0 and fwdSpeed > -0.1 then
		blocked = 1.0 - clamp(fwdSpeed/0.5, 0.0, 1.0)
	end
	robot.blocked = robot.blocked * 0.95 + blocked * 0.05

	--Always blocked if fall is detected
	if sensor.detectFall > 0 then
		robot.blocked = 1.0
	end

	--Evaluate mass every frame since robots can break
	robot.mass = 0
	local bodies = FindBodies()
	for i=1, #bodies do
		robot.mass = robot.mass + GetBodyMass(bodies[i])
	end
	
	robot.bodyCenter = TransformToParentPoint(robot.transform, GetBodyCenterOfMass(robot.body))
	robot.navigationCenter = TransformToParentPoint(robot.transform, Vec(0, -hover.distTarget, 0))

	--Handle break all
	robot.breakAllTimer = math.max(0.0, robot.breakAllTimer - dt)
	if not robot.breakAll and robot.breakAllTimer > 0.0 then
		for i=1, #robot.allShapes do
			SetTag(robot.allShapes[i], "breakall")
		end
		robot.breakAll = true
	end
	if robot.breakAll and robot.breakAllTimer <= 0.0 then
		for i=1, #robot.allShapes do
			RemoveTag(robot.allShapes[i], "breakall")
		end
		robot.breakAll = false
	end
	
	--Distance and direction to player
	local pp = VecAdd(GetPlayerTransform().pos, Vec(0, 1, 0))
	local d = VecSub(pp, robot.bodyCenter)
	robot.distToPlayer = VecLength(d)
	robot.dirToPlayer = VecScale(d, 1.0/robot.distToPlayer)
	
	--Sense player if player is close and there is nothing in between
	robot.canSensePlayer = false
	if robot.distToPlayer < 3.0 then
		rejectAllBodies(robot.allBodies)
		if not QueryRaycast(robot.bodyCenter, robot.dirToPlayer, robot.distToPlayer) then
			robot.canSensePlayer = true
		end
	end

	--Robot body sounds
	if robot.enabled and hover.contact > 0 then
		local vol
		vol = clamp(VecLength(GetBodyVelocity(robot.body)) * 0.4, 0.0, 1.0)
		if vol > 0 then
			PlayLoop(walkLoop, robot.transform.pos, vol)
		end

		vol = clamp(VecLength(GetBodyAngularVelocity(robot.body)) * 0.4, 0.0, 1.0)
		if vol > 0 then
			PlayLoop(turnLoop, robot.transform.pos, vol)
		end
	end
end


------------------------------------------------------------------------


hover = {}
hover.hitBody = 0
hover.contact = 0.0
hover.distTarget = 1.1
hover.distPadding = 0.3
hover.timeSinceContact = 0.0





------------------------------------------------------------------------



------------------------------------------------------------------------


feet = {}

function feetInit()
	--+--+print("feetInit")
	local f = FindBodies("foot")
	for i=1, #f do
		local foot = {}
		foot.body = f[i]
		local t = GetBodyTransform(foot.body)
		local rayOrigin = TransformToParentPoint(t, Vec(0, 0.9, 0))
		local rayDir = TransformToParentVec(t, Vec(0, -1, 0))

		foot.lastTransform = TransformCopy(t)
		foot.targetTransform = TransformCopy(t)
		foot.candidateTransform = TransformCopy(t)
		foot.worldTransform = TransformCopy(t)
		foot.stepAge = 1
		foot.stepLifeTime = 1
		foot.localRestTransform = TransformToLocalTransform(robot.transform, t)
		foot.localTransform = TransformCopy(foot.localRestTransform)
		foot.rayOrigin = TransformToLocalPoint(robot.transform, rayOrigin)
		foot.rayDir = TransformToLocalVec(robot.transform, rayDir)
		foot.rayDist = hover.distTarget + hover.distPadding
		foot.contact = true
		local mass = GetBodyMass(foot.body)
		foot.linForce = 20 * mass
		foot.angForce = 1 * mass
		local linScale, angScale = getTagParameter2(foot.body, "force", 1.0)
		foot.linForce = foot.linForce * linScale
		foot.angForce = foot.angForce * angScale
		feet[i] = foot

	end
end



------------------------------------------------------------------------




------------------------------------------------------------------------


sensor = {}
sensor.blocked = 0
sensor.blockedLeft = 0
sensor.blockedRight = 0
sensor.detectFall = 0


function sensorInit()

end



------------------------------------------------------------------------


head = {}
head.body = 0
head.eye = 0
head.dir = Vec(0,0,-1)
head.lookOffset = 0
head.lookOffsetTimer = 0
head.canSeePlayer = false
head.lastSeenPos = Vec(0,0,0)
head.timeSinceLastSeen = 999
head.seenTimer = 0
head.alarmTimer = 0
head.alarmTime = 2.0
head.aim = 0	-- 1.0 = perfect aim, 0.0 = will always miss player. This increases when robot sees player based on config.aimTime


function headInit()
	--+--+print("headInit")
	head.body = FindBody("head")
	head.eye = FindLight("eye")
	head.joint = FindJoint("head")
	head.alarmTime = getTagParameter(head.eye, "alarm", 2.0)
end


------------------------------------------------------------------

------------------------------------------------------------------------

navigation = {}
navigation.state = "done"
navigation.path = {}
navigation.target = Vec()
navigation.hasNewTarget = false
navigation.resultRetrieved = true
navigation.deviation = 0		-- Distance to path
navigation.blocked = 0
navigation.unblockTimer = 0		-- Timer that ticks up when blocked. If reaching limit, unblock kicks in and timer resets
navigation.unblock = 0			-- If more than zero, navigation is in unblock mode (reverse direction)
navigation.vertical = 0
navigation.thinkTime = 0
navigation.timeout = 1
navigation.lastQueryTime = 0
navigation.timeSinceProgress = 0


function init()
	--+--+print("init")



	configInit()
	robotInit()



	oneHitSniper = GetBool("savegame.mod.destructible-bot.oneHitSniper")
	robotAttackOneHit = GetBool("savegame.mod.destructible-bot.robotAttackOneHit")
	robotOutline = GetBool("savegame.mod.destructible-bot.robotOutline")
	robotBetterSensors = GetBool("savegame.mod.destructible-bot.robotBetterSensors")
	robotHealth = GetFloat("savegame.mod.destructible-bot.robotHealth")
	oneHitEverything = GetBool("savegame.mod.destructible-bot.oneHitEverything")
	autoremove = GetBool("savegame.mod.destructible-bot.autoremove")
	
	--[[  
		Uncomment if playing in career mode
	oneHitSniper = false
	robotAttackOneHit = true
	robotOutline = false
	robotBetterSensors = true
	robotHealth = 1
	]]--


	timetodie = math.random(0,4) - math.random()
	drawHealth = 0
	uiooffset = 0
	rectHeightFeet = {}
	rectHeightLeg = {}

	-- Fall Damage vars
	knockback = 0
	accumulator = 0
	robotVel = 0
	robotVelPrev = 0
	accumSpeed = 0
	accumulator = 0
	prevAccumSpeed = 0
	maxdif = 0
	dif = 0
	spankmax = 0
	spank = 0
	maxdiffall = 0
	delay = 0
	timeuntilremoved = 300
	spankdelay = 0
	spankreset = false
	gotspanked = false


	totalhealth = 0 
	attackmultiplier = 1
	for i=1, #robot.allShapes do
		SetTag(robot.allShapes[i], "unbreakable")

		SetTag(robot.allShapes[i],'shapeHealth',(math.sqrt(GetShapeVoxelCount(robot.allShapes[i]))*4)*robotHealth)
		SetTag(robot.allShapes[i],'shapeHealthInit',(math.sqrt(GetShapeVoxelCount(robot.allShapes[i]))*4)*robotHealth)
		addThis = math.sqrt(GetShapeVoxelCount(robot.allShapes[i]))*4*robotHealth
		if HasTag(GetShapeBody(robot.allShapes[i]),"head") then
			robothead = robot.allShapes[i]
			robotheadcount = GetShapeVoxelCount(robot.allShapes[i])
			SetTag(robot.allShapes[i],'shapeHealth',math.sqrt(GetShapeVoxelCount(robot.allShapes[i]))*5.5*robotHealth+150)
			SetTag(robot.allShapes[i],'shapeHealthInit',math.sqrt(GetShapeVoxelCount(robot.allShapes[i]))*5.5*robotHealth+150)
			addThis = math.sqrt(GetShapeVoxelCount(robot.allShapes[i]))*5.5*robotHealth+150
		end
		if HasTag(GetShapeBody(robot.allShapes[i]),"foot") then
			SetTag(robot.allShapes[i],'shapeHealth',math.sqrt(GetShapeVoxelCount(robot.allShapes[i]))*10*robotHealth)
			SetTag(robot.allShapes[i],'shapeHealthInit',math.sqrt(GetShapeVoxelCount(robot.allShapes[i]))*10*robotHealth)
			SetTag(GetShapeBody(robot.allShapes[i]),"brokenfoot","false")
			addThis = math.sqrt(GetShapeVoxelCount(robot.allShapes[i]))*10*robotHealth
		end
		if HasTag(GetShapeBody(robot.allShapes[i]),"leg") then
			SetTag(robot.allShapes[i],'shapeHealth',math.sqrt(GetShapeVoxelCount(robot.allShapes[i]))*7*robotHealth)
			SetTag(robot.allShapes[i],'shapeHealthInit',math.sqrt(GetShapeVoxelCount(robot.allShapes[i]))*7*robotHealth)
			addThis = math.sqrt(GetShapeVoxelCount(robot.allShapes[i]))*7*robotHealth
		end
		if HasTag(GetShapeBody(robot.allShapes[i]),"aim") then
			SetTag(robot.allShapes[i],'shapeHealth',math.sqrt(GetShapeVoxelCount(robot.allShapes[i]))*13*robotHealth)
			SetTag(robot.allShapes[i],'shapeHealthInit',math.sqrt(GetShapeVoxelCount(robot.allShapes[i]))*13*robotHealth)
			addThis = math.sqrt(GetShapeVoxelCount(robot.allShapes[i]))*13*robotHealth
		end
		totalhealth = totalhealth + addThis
	end

	SetTag(FindBody('head'),'alive')

	totalhealth = totalhealth
	hitcounter = 0
	initialtotal = totalhealth

	patrolLocations = FindLocations("patrol",true)
	shootSound = LoadSound("tools/gun0.ogg", 8.0)
	rocketSound = LoadSound("tools/launcher0.ogg", 7.0)
	local nomDist = 7.0
	if config.stepSound == "s" then nomDist = 5.0 end
	if config.stepSound == "l" then nomDist = 9.0 end
	stepSound = LoadSound("robot/step-" .. config.stepSound .. "0.ogg", nomDist)
	headLoop = LoadLoop("robot/head-loop.ogg", 7.0)
	turnLoop = LoadLoop("robot/turn-loop.ogg", 7.0)
	walkLoop = LoadLoop("robot/walk-loop.ogg", 7.0)
	rollLoop = LoadLoop("robot/roll-loop.ogg", 7.0)
	chargeLoop = LoadLoop("robot/charge-loop.ogg", 8.0)
	alertSound = LoadSound("robot/alert.ogg", 9.0)
	huntSound = LoadSound("robot/hunt.ogg", 9.0)
	idleSound = LoadSound("robot/idle.ogg", 9.0)
	fireLoop = LoadLoop("tools/blowtorch-loop.ogg")
	disableSound = LoadSound("robot/disable0.ogg")
end



function update(dt)
	--+--+print("update")

	if robot.deleted then
		return
	else
		if autoremove then
			local dead = 0
			for i=1, #robot.allShapes do 
				if HasTag(robot.allShapes[i],'toberemoved') then
					triggertimer = true
				end
			end
			if timeuntilremoved < 0 and false == IsBodyVisible(GetShapeBody(robot.allShapes[i]),40,true) then
				Delete(robot.allShapes[i])
				if IsHandleValid(robot.body) then
					for i=1, #robot.allBodies do
						Delete(robot.allBodies[i])
					end
					for i=1, #robot.allJoints do
						Delete(robot.allJoints[i])
					end
				end
			end
				
			if triggertimer == true then
				deadRobots = FindBodies("head",true)
				for i=1, #deadRobots do
					if HasTag(deadRobots[i],'dead') then
						dead = dead + 1
					end
				end
				timeuntilremoved = timeuntilremoved - (dt*dead) 
			end
		end
	end

	if robot.deleted then 
		return
	else 
		if not IsHandleValid(robot.body) then
			for i=1, #robot.allBodies do
				Delete(robot.allBodies[i])
			end
			for i=1, #robot.allJoints do
				Delete(robot.allJoints[i])
			end
			robot.deleted = true
		end
	end
	robotUpdate(dt)

	if not robot.enabled then
		return
	end
	if IsPointInWater(robot.bodyCenter) or 0 > totalhealth or IsShapeBroken(robothead) and GetShapeVoxelCount(robothead)/robotheadcount < 0.88   then
		timetodie = timetodie - dt
		print("dead")
		SetTag(robot.body, "shutdown")
		if timetodie < 0 or timetodie >= 2 and timetodie <= 6 or GetShapeVoxelCount(robothead)/robotheadcount then
			PlaySound(disableSound, robot.bodyCenter)
			for i=1, #robot.allShapes do
				SetShapeEmissiveScale(robot.allShapes[i], 0)
				SetTag(robot.allShapes[i], "toberemoved")
			end
			print("even more dead")
			SetTag(FindBody('head'),'dead')
			SetTag(robot.body, "disabled")
			RemoveTag(robot.body, "shutdown")
			RemoveTag(FindBody('head'),'alive')
			RemoveTag(FindBody('head'),'disarmed')
			SetInt("savegame.mod.destructible-bot.deletedbots",GetInt("savegame.mod.destructible-bot.deletedbots")+1)
			robot.enabled = false
			print("finish of dead")
		end
	end
end



function tick(dt)
end


function decreaseHealthShot(damagebeforecalc,shape)
	if oneHitEverything == true then
		totalhealth = totalhealth -( totalhealth +1)
		SetTag(shape,'shapeHealth',tonumber(GetTagValue(shape,'shapeHealth'))-(GetTagValue(shape,'shapeHealth')+1))
		RemoveTag(shape,"unbreakable")
	end
end

  function hitByShot(strength, pos, dir)
	local hit, point, n, shape = QueryClosestPoint(pos, 0.1)
	if VecDist(pos, robot.bodyCenter) < 3 then
		local hit, point, n, shape = QueryClosestPoint(pos, 0.1)
		if hit and HasTag(shape,'shapeHealth') then
		if GetString("game.player.tool") == "gun" then
		decreaseHealthShot(35,shape)
		end
		end
	end
end


---------------------------------------------------------------------------------

function handleCommand(cmd)
	print("handleCommand")
	words = splitString(cmd, " ")
	if robot.deleted == false then
		if #words == 8 then
			if words[1] == "shot" then
				local strength = tonumber(words[2])
				local x = tonumber(words[3])
				local y = tonumber(words[4])
				local z = tonumber(words[5])
				local dx = tonumber(words[6])
				local dy = tonumber(words[7])
				local dz = tonumber(words[8])

				hitByShot(strength, Vec(x,y,z), Vec(dx,dy,dz))
			end
		end
	end
	print("end of handleCommand")
end
